SystemOrganization addCategory: #'ContextS2-Core'!

!Array methodsFor: '*contexts2-core' stamp: 'ag 1/27/2012 16:48:52.707'!
withLayersDo: aBlock
	"Memorize currently active layers, add the new ones and restore the state afterwards.
	The layer-name-symbols are extracted from the receiver, in case real CSLayer-object are stored here.
	TODO in future, change internal layer-management to only use real layer-objects."
	
	| layerNames |
	layerNames := self collect: #layerName.
	^ Processor activeProcess csInfo
		withLayers: (Processor activeProcess csInfo activeLayers copyWithoutAll: layerNames), layerNames
		do: aBlock! !

!Array methodsFor: '*contexts2-core' stamp: 'ag 12/20/2011 21:31'!
withoutLayersDo: aBlock
	"Memorize currently active layers, remove some and restore the state afterwards."
	
	^ Processor activeProcess csInfo
		withLayers: (Processor activeProcess csInfo activeLayers copyWithoutAll: self)
		do: aBlock! !

!CompiledMethod methodsFor: '*contexts2-core' stamp: 'ag 2/1/2012 16:19:51.427'!
asPartialMethod
	
	| layer |
	layer := self layersPragma
		ifNotNilDo: [ :pragma |
			pragma arguments size == 1 ifFalse: [ self error: 'Illegal ContextS2-Pragma: Need only one argument.' ].
			layer := pragma argumentAt: 1.
			(layer isSymbol | (layer isArray and: [ layer allSatisfy: #isSymbol ]))
				ifFalse: [ self error: 'Illegal ContextS2-Pragma: Need Symbol or Array of Symbols as parameter.' ].
			layer isArray ifTrue: [ layer ] ifFalse: [ Array with: layer ] ].
	^ CSPartialMethod
				on: layer
				forMethod: self! !

!CompiledMethod methodsFor: '*contexts2-core' stamp: 'tfel 2/23/2012 17:46:49.222'!
asSymbol

	^ self selector! !

!CompiledMethod methodsFor: '*contexts2-core' stamp: 'ag 2/1/2012 16:19:44.324'!
layersPragma
	
	^ self pragmas
		detect: [ :each | each keyword == CSLayeredMethod pragmaKeyword ]
		ifNone: [ nil ]! !

MethodReference subclass: #CSMethodReference
	instanceVariableNames: 'method'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ContextS2-Core'!

!CSMethodReference class methodsFor: 'instance creation' stamp: 'tfel 2/23/2012 18:54:11.379'!
onMethod: aMethod

	^ (self class: aMethod methodClass selector: aMethod selector)
		method: aMethod;
		yourself! !

!CSMethodReference methodsFor: 'monticello' stamp: 'tfel 2/23/2012 19:06:54.797'!
asMethodDefinition
	"Override, unification happens based on class and selector in MC"
	^ (MCMethodDefinition forMethodReference: self)
		selector: self method asSymbol;
		yourself! !

!CSMethodReference methodsFor: 'monticello' stamp: 'tfel 2/23/2012 18:42:27.33'!
compiledMethod

	^ self method! !

!CSMethodReference methodsFor: 'accessing' stamp: 'tfel 2/23/2012 18:40:15.256'!
method
	"Answer the value of method"

	^ method! !

!CSMethodReference methodsFor: 'accessing' stamp: 'tfel 2/23/2012 18:40:15.277'!
method: anObject
	"Set the value of method"

	method := anObject! !

!CSMethodReference methodsFor: 'monticello' stamp: 'tfel 2/23/2012 18:55:02.754'!
source

	^ self method getSource asString! !

!CSMethodReference methodsFor: 'packageinfo-base' stamp: 'tfel 2/23/2012 18:55:09.204'!
sourceCode

	^ self source asString! !

!Parser class methodsFor: '*contexts2-core-override-accessing' stamp: 'ag 12/20/2011 21:39'!
primitivePragmaSelectors
	"Answer the selectors of pragmas that specify VM primitives.
	 Needed for compile and decomple."
	^#(primitive:
		primitive:error:
		primitive:error:module:
		primitive:module:
		primitive:module:error:
		), CSLayeredMethod pragmaKeyword! !

!Parser methodsFor: '*contexts2-core' stamp: 'ag 12/20/2011 21:40'!
layer: aCollectionOfLayers
	^ 0! !

Browser subclass: #CSBrowser
	instanceVariableNames: 'selectedLayerSymbol'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ContextS2-Core'!

!CSBrowser class methodsFor: 'constants' stamp: 'tfel 2/23/2012 16:30:14.747'!
allLayersMarker
	
	^ #'-- all --'! !

!CSBrowser class methodsFor: 'constants' stamp: 'tfel 2/23/2012 16:30:09.869'!
baseMethodsMarker
	
	^ #'-- base methods --'! !

!CSBrowser class methodsFor: 'class initialization' stamp: 'tfel 2/23/2012 18:15:51.225'!
initialize

	super initialize.
	TheWorldMenu registerOpenCommand: {'ContextS Browser'. {self. #open}}.! !

!CSBrowser class methodsFor: 'constants' stamp: 'tfel 2/23/2012 16:30:04.976'!
nonLayeredMarker
	
	^ #'-- non-layered --'! !

!CSBrowser methodsFor: 'message list' stamp: 'tfel 2/27/2012 18:39:22.284'!
allLayersAndMethodsFrom: aList

	^ Array streamContents: [:stream |
		aList do: [:method |
			method isContextSMethod
					ifTrue: [
						method basePartialMethod ifNotNilDo: [:m | stream nextPut: m].
						stream nextPutAll: method partialMethods]
					ifFalse: [stream nextPut: method]]]! !

!CSBrowser methodsFor: 'message list' stamp: 'tfel 2/27/2012 18:38:48.296'!
baseMethodsFrom: aList

	^ (aList
		select: [:method | method isContextSMethod]
		thenCollect: [:method | method basePartialMethod])
		reject: [:method | method isNil]! !

!CSBrowser methodsFor: 'Tools' stamp: 'tfel 2/28/2012 16:06:36.843'!
browseVersions

	^ [super browseVersions]
		on: Error
		do: [:e | self inform: 'Sorry. Only actual methods can show versions'].! !

!CSBrowser methodsFor: 'toolbuilder' stamp: 'tfel 2/23/2012 16:02:42.692'!
buildContextLayerListWith: builder
	
	^ builder pluggableListSpec new
		model: self;
		list: #contextLayerList;
		getIndex: #selectedLayerIndex;
		setIndex: #selectedLayerIndex:;
		menu: #contextLayerMenu:;
		keyPress: #arrowKey:from:;
		yourself! !

!CSBrowser methodsFor: 'toolbuilder' stamp: 'tfel 2/23/2012 17:56:37.637'!
buildWindowWith: builder specs: aSpecList
	| paneSpecs codeSpec slimPaneSpecs layerPaneSpec layerPaneFrame |
	paneSpecs := aSpecList first: 5.
	codeSpec := aSpecList last.
	slimPaneSpecs := self compressWidthOfSpecs: paneSpecs from: 1/4 to: 1/5.
	layerPaneFrame := slimPaneSpecs first key copy
							leftFraction: 0.8;
							rightFraction: 1;
							yourself.
	layerPaneSpec := layerPaneFrame -> [self buildContextLayerListWith: builder].
	self when: #classSelectionChanged send: #resetLayers to: self.
	^ super
		buildWindowWith: builder
		specs: slimPaneSpecs, {layerPaneSpec}, {codeSpec}! !

!CSBrowser methodsFor: 'toolbuilder' stamp: 'tfel 2/23/2012 16:46:17.613'!
compressWidthOfSpecs: specs from: aFraction to: anotherFraction

	^ specs collect: [:assoc || frame |
		frame := assoc key isRectangle
					ifTrue: [LayoutFrame fractions: assoc key] ifFalse: [assoc key].
		frame rightFraction: frame rightFraction / aFraction * anotherFraction.
		frame leftFraction: frame leftFraction / aFraction * anotherFraction.
		frame -> assoc value].
! !

!CSBrowser methodsFor: 'layer list' stamp: 'tfel 2/23/2012 16:04:58.234'!
contextLayerList
	
	^ self selectedClassOrMetaClass
		ifNil: [#()]
		ifNotNilDo: [:class |
			{ self class allLayersMarker.
			self class nonLayeredMarker.
			self class baseMethodsMarker },
			self selectedClassOrMetaClass allInvolvedLayers asArray ]! !

!CSBrowser methodsFor: 'layer list' stamp: 'tfel 2/23/2012 18:01:34.413'!
contextLayerMenu: aMenu
	
	(self menuHook: aMenu named: #messageCategoryMenu shifted: false).
	^ aMenu
		labels: 'delete layer'
		lines: #()
		selections: #(deleteSelectedLayer)! !

!CSBrowser methodsFor: 'initialize-release' stamp: 'tfel 2/23/2012 16:16:16.924'!
defaultBrowserTitle
	
	^ 'Context Browser'! !

!CSBrowser methodsFor: 'layer functions' stamp: 'tfel 2/23/2012 18:05:09.168'!
deleteSelectedLayer

	self inform: 'Not implemented'.! !

!CSBrowser methodsFor: 'user interface' stamp: 'tfel 2/28/2012 10:31:38.455'!
initialExtent
	| baseExtent |
	baseExtent := super initialExtent.
	^ (baseExtent x / 4 * 5)@(baseExtent y)! !

!CSBrowser methodsFor: 'message list' stamp: 'tfel 2/23/2012 17:46:28.392'!
messageList
	
	^ self messageObjectsList collect: #asSymbol! !

!CSBrowser methodsFor: 'message list' stamp: 'tfel 2/23/2012 17:41:37.844'!
messageObjectsList
	| list |
	list := super messageList collect: [:selector |
					self selectedClassOrMetaClass >> selector].
	^ self selectedLayerSymbol
		caseOf: {[self class nonLayeredMarker] -> [self nonLayeredMessagesFrom: list].
	 				[self class allLayersMarker] -> [self allLayersAndMethodsFrom: list].
	 				[self class baseMethodsMarker] -> [self baseMethodsFrom: list]}
		otherwise: [self messagesInLayer: self selectedLayerSymbol from: list]! !

!CSBrowser methodsFor: 'message list' stamp: 'tfel 2/23/2012 17:42:34.234'!
messagesInLayer: aSymbol from: aList

	^ (Array streamContents: [:stream | (aList
		select: [:method | method isContextSMethod]
		thenCollect: [:method | stream nextPutAll: method partialMethods])])
		select: [:method | method layers includes: self selectedLayerSymbol]! !

!CSBrowser methodsFor: 'message list' stamp: 'tfel 2/23/2012 17:45:25.34'!
nonLayeredMessagesFrom: aList

	^ aList select: [:method | method isContextSMethod not]! !

!CSBrowser methodsFor: 'message functions' stamp: 'tfel 2/23/2012 18:11:46.858'!
removeMessage
	"Changed to delete only one partial method for ContextS methods."
	| method confirmation |
	self messageListIndex = 0 | self okToChange not ifTrue: [^ self].
	method := self messageObjectsList at: self messageListIndex.
	method isContextSMethod ifFalse: [^ super removeMessage].

	confirmation := self systemNavigation 
							confirmRemovalOf: method selector
							on: self selectedClassOrMetaClass.
	confirmation = 3 ifTrue: [^ self].
	
	(self selectedClassOrMetaClass >> method selector) removePartialMethod: method.
	self changed: #messageList.
	self messageListIndex: 0.
	
	self setClassOrganizer.
	"In case organization not cached"
	confirmation = 2 ifTrue: [self systemNavigation browseAllCallsOn: method selector].! !

!CSBrowser methodsFor: 'layer list' stamp: 'tfel 2/23/2012 17:57:30.108'!
resetLayers
	
	self selectedLayerIndex: 0.  ! !

!CSBrowser methodsFor: 'layer list' stamp: 'tfel 2/23/2012 16:07:53.324'!
selectedLayerIndex
	
	^ self contextLayerList indexOf: self selectedLayerSymbol! !

!CSBrowser methodsFor: 'layer list' stamp: 'tfel 2/23/2012 16:50:10.884'!
selectedLayerIndex: anInteger
	
	self selectedLayerSymbol: (self contextLayerList at: anInteger ifAbsent: []).
	self
		changed: #selectedLayerIndex;
		changed: #contextLayerList;
		changed: #messageCategoryList;
		changed: #relabel;
		contentsChanged;
		messageCategoryListIndex: 1. "Select the --all-- category, deselect messages."! !

!CSBrowser methodsFor: 'accessing' stamp: 'tfel 2/23/2012 16:06:26.772'!
selectedLayerSymbol
	"This returns the name of the layer (as symbol), not the layer object."
	
	^ selectedLayerSymbol ifNil: [selectedLayerSymbol := self class allLayersMarker]! !

!CSBrowser methodsFor: 'accessing' stamp: 'tfel 2/23/2012 16:07:40.398'!
selectedLayerSymbol: anObject
	"Set the value of selectedLayerSymbol"

	selectedLayerSymbol := anObject! !

!CSBrowser methodsFor: 'message list' stamp: 'tfel 2/23/2012 17:49:01.638'!
selectedMessage

	^ (super selectedMessage isEmpty and: [self messageListIndex = 0])
		ifTrue: [^ '']
		ifFalse: [(self messageObjectsList at: self messageListIndex) getSource]
! !

Browser subclass: #CSContextBrowser
	instanceVariableNames: 'selectedLayerIndex contextLayerList messageCategoryList messageList'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ContextS2-Core'!

!CSContextBrowser class methodsFor: 'constants' stamp: 'ag 1/22/2012 15:22:53.043'!
allLayersMarker
	
	^ #'-- all --'! !

!CSContextBrowser class methodsFor: 'constants' stamp: 'ag 1/20/2012 22:16:40.498'!
baseMethodsMarker
	
	^ #'-- base methods --'! !

!CSContextBrowser class methodsFor: 'constants' stamp: 'ag 1/22/2012 10:34:34.381'!
nonLayeredMarker
	
	^ #'-- non-layered --'! !

!CSContextBrowser methodsFor: 'ui-layers' stamp: 'ag 1/22/2012 18:20:57.702'!
buildContextLayerListWith: builder
	
	^ builder pluggableListSpec new
		model: self;
		list: #contextLayerList;
		getIndex: #selectedLayerIndex;
		setIndex: #selectedLayerIndex:;
		menu: #contextLayerMenu:;
		keyPress: #arrowKey:from:;
		yourself! !

!CSContextBrowser methodsFor: 'ui' stamp: 'tfel 2/23/2012 14:43:05.903'!
buildMessageCategoryBrowserEditString: aString
	"Copied from superclass, changed to integrate with other functionality of this class (Correctly select the index of the selected message category)."

	"Create and schedule a message category browser for the currently 
	selected	 message category. The initial text view contains the characters 
	in aString."
	"wod 6/24/1998: set newBrowser classListIndex so that it works whether the
	receiver is a standard or a Hierarchy Browser."

	| newBrowser |
	self messageCategoryListIndex ~= 0
		ifTrue: 
			[newBrowser := Browser new.
			newBrowser metaClassIndicated: metaClassIndicated.			
			newBrowser systemCategoryListIndex: (newBrowser systemCategoryList indexOf: self selectedSystemCategoryName).
			newBrowser classListIndex: (newBrowser classList indexOf: self selectedClassName).
			newBrowser messageCategoryListIndex: (newBrowser messageCategoryList indexOf: self selectedMessageCategoryName).
			self selectedMessageName ifNotNil: [
				newBrowser messageListIndex: (newBrowser messageList indexOf: self selectedMessageName) ].

			self class openBrowserView: (newBrowser openMessageCatEditString: aString)
				label: 'Message Category Browser (' , 
						newBrowser selectedClassOrMetaClassName , ')']! !

!CSContextBrowser methodsFor: 'initialize-release' stamp: 'ag 1/31/2012 23:00:51.057'!
changed: aSymbol
	"TODO maybe this is not the right way to get the changed-event of the reicever itself..."
	
	aSymbol caseOf: {
		[ #messageList ] -> [ messageList := nil ].
		[ #messageCategoryList ] -> [ messageCategoryList := nil ].
		[ #contextLayerList ] -> [ contextLayerList := nil ].
	} otherwise: [].
	super changed: aSymbol.! !

!CSContextBrowser methodsFor: 'ui' stamp: 'tfel 2/23/2012 14:43:29.32'!
classListIndex: aNumber
	
	selectedLayerIndex := 0. "Set the variable first, do all the real updates after the super-call."
	self messageCategoryListIndex: 0.
	super classListIndex: aNumber.
	self selectedLayerIndex: 0.! !

!CSContextBrowser methodsFor: 'ui-messages' stamp: 'ag 1/31/2012 15:29:58.83'!
completeMessageList
	
	| messages |
	messages := SortedCollection sortBlock: [ :mt1 :mt2 |
		(mt1 isSymbol ifTrue: [ mt1 ] ifFalse: [ mt1 selector ]) <= (mt2 isSymbol ifTrue: [ mt2 ] ifFalse: [ mt2 selector ]) ].
	(self selectedClassOrMetaClass ifNil: [ ^ #() ]) methodDict
		valuesDo: [ :method |
			method selector ifNotEmpty: [
				(method isKindOf: CSLayeredMethod)
					ifTrue: [ messages addAll: method allPartialMethods ]
					ifFalse: [ messages add: method selector ] ] ].
	^ messages! !

!CSContextBrowser methodsFor: 'ui-layers' stamp: 'tfel 2/23/2012 14:41:46.439'!
contextLayerList
	
	^ contextLayerList ifNil: [
		contextLayerList := self classListIndex = 0
			ifTrue: [ Array new ] "No class selected"
			ifFalse: [ 
				{ self class allLayersMarker. self class nonLayeredMarker. self class baseMethodsMarker },
				self selectedClassOrMetaClass allInvolvedLayers asArray ] ]! !

!CSContextBrowser methodsFor: 'ui-layers' stamp: 'ag 1/22/2012 18:25:48.119'!
contextLayerMenu: aMenu
	
	(self menuHook: aMenu named: #messageCategoryMenu shifted: false) ifTrue: [ ^ aMenu ].
	^ aMenu
		labels: 'delete layer'
		lines: #()
		selections: #(deleteSelectedLayer)! !

!CSContextBrowser methodsFor: 'initialize-release' stamp: 'ag 1/12/2012 12:06'!
defaultBrowserTitle
	
	^ 'Context Browser'! !

!CSContextBrowser methodsFor: 'ui-actions' stamp: 'ag 1/20/2012 21:42:35.885'!
deleteSelectedLayer
	"When the layer is deleted, we delete all partial methods on it.
	If the partial method was the last one in its LayeredMethod, we delete the LayeredMethod <??>"
	
	self selectedLayer delete.
	self selectedLayerIndex: 0.! !

!CSContextBrowser methodsFor: 'initialize-release' stamp: 'ag 1/31/2012 15:32:27.19'!
initialize
	
	super initialize.
	selectedLayerIndex := 0.! !

!CSContextBrowser methodsFor: 'ui' stamp: 'ag 1/31/2012 23:05:22.837'!
messageCategoryList
	
	^ messageCategoryList ifNil: [
		messageCategoryList := SortedCollection with: ClassOrganizer allCategory.
		(self selectedClassOrMetaClass ifNil: [ ^ #() ]) methodDict keysAndValuesDo: [ :key :value |
			(self selectedLayer = self class allLayersMarker
			or: [
			(self selectedLayer = self class nonLayeredMarker)
				ifTrue: [ value class ~= CSLayeredMethod ]
				ifFalse: [
					value class == CSLayeredMethod and: [
						(self selectedLayer = self class baseMethodsMarker
							ifTrue: [ value hasBaseMethod ]
							ifFalse: [ value allInvolvedLayers includes: self selectedLayer ]) ] ] ]
			) ifTrue: [
				(self classOrMetaClassOrganizer categoryOfElement: key)
					ifNotNilDo: [ :category | messageCategoryList addIfNotPresent: category ] ] ].
		messageCategoryList ]! !

!CSContextBrowser methodsFor: 'ui-messages' stamp: 'ag 1/31/2012 22:54:34.758'!
messageList
	"Select those (partial) methods, that match both the selected category and selected layer."
	
	^ messageList ifNil: [
		messageList := self completeMessageList select: [ :method |
			"First filter the category"
			(self selectedMessageCategoryName isNil
				or: [ self selectedMessageCategoryName = ClassOrganizer allCategory ]
				or: [ (self classOrMetaClassOrganizer categoryOfElement: method asSymbol) = self selectedMessageCategoryName ])
			and: [
			
			"Then the layer"
			self selectedLayer = self class allLayersMarker
			or: [
			(self selectedLayer = self class nonLayeredMarker)
				ifTrue: [ (method isKindOf: CSMethodObject) not ]
				ifFalse: [
					(method isKindOf: CSMethodObject) and: [
						(self selectedLayer = self class baseMethodsMarker)
							ifTrue: [ method isBase ]
							ifFalse: [ method layers includes: self selectedLayer ] ] ] ] ] ] ]! !

!CSContextBrowser methodsFor: 'initialize-release' stamp: 'ag 1/31/2012 15:33:37.095'!
metaClassIndicated: aBool
	"This value is set when one of the instance/class buttons is clicked."
	
	selectedLayerIndex := 0.
	super metaClassIndicated: aBool.
	self changed: #contextLayerList.! !

!CSContextBrowser methodsFor: 'initialize-release' stamp: 'ag 1/22/2012 18:09:56.262'!
openEditString: aString
	"Create a pluggable version of all the views for a Browser, including views and controllers."
	"CSContextBrowser fullOnClass: Browser"
	
	| builder max |
	builder := ToolBuilder default.
	max := self wantsOptionalButtons ifTrue: [ 0.42 ] ifFalse: [ 0.5 ].
	^ self buildWindowWith: builder specs: {
		(0@0 corner: 0.2@max) -> [self buildSystemCategoryListWith: builder].
		(0.2@0 corner: 0.4@(max-0.08)) -> [self buildClassListWith: builder].
		(0.2@(max-0.08) corner: 0.4@max) -> [self buildSwitchesWith: builder].
		(0.6@0 corner: 0.8@max) -> [ self buildMessageCategoryListWith: builder ].
		(0.4@0 corner: 0.6@max) -> [self buildContextLayerListWith: builder].
		(0.8@0 corner: 1@max) -> [self buildMessageListWith: builder].
		(0@max corner: 1@1) -> [self buildCodePaneWith: builder]
	}! !

!CSContextBrowser methodsFor: 'ui-messages' stamp: 'tfel 2/23/2012 14:44:06.225'!
removeMessage
	"Copied from super-class, changed to delete only one partial method."
	
	| method confirmation |
	self messageListIndex = 0 | self okToChange not ifTrue: [^ self].
	method := self selectedMethod.
	method isSymbol ifTrue: [ ^ super removeMessage ].
	confirmation := self systemNavigation confirmRemovalOf: method selector on: self selectedClassOrMetaClass.
	confirmation = 3 ifTrue: [^ self].
	
	(self selectedClassOrMetaClass >> method selector) removePartialMethod: method.
	self changed: #messageList.
	self messageListIndex: 0.
	
	self setClassOrganizer.
	"In case organization not cached"
	confirmation = 2 ifTrue: [self systemNavigation browseAllCallsOn: method selector]! !

!CSContextBrowser methodsFor: 'ui-layers' stamp: 'ag 1/22/2012 14:46:04.694'!
selectedLayer
	"This returns the name of the layer (as symbol), not the layer object."
	
	^ self selectedLayerIndex = 0
		ifTrue: [ self class allLayersMarker ]
		ifFalse: [ self contextLayerList at: self selectedLayerIndex ]! !

!CSContextBrowser methodsFor: 'ui-layers' stamp: 'ag 1/12/2012 12:45'!
selectedLayerIndex
	^ selectedLayerIndex! !

!CSContextBrowser methodsFor: 'ui-layers' stamp: 'tfel 2/23/2012 14:43:49.154'!
selectedLayerIndex: anObject
	
	selectedLayerIndex := anObject.
	self messageCategoryListIndex: 1. "First just reset the variable, later do the real updates."
	self messageListIndex: 0.
	self
		changed: #selectedLayerIndex;
		changed: #contextLayerList;
		changed: #messageCategoryList;
		changed: #relabel;
		contentsChanged;
		messageCategoryListIndex: 1. "Select the --all-- category, deselect messages."! !

!CSContextBrowser methodsFor: 'ui-messages' stamp: 'ag 1/22/2012 23:08:24.183'!
selectedMessage
	"If the selected message is a normal CompiledMethods (the object then is the selector-symbol), use the super-version.
	Else, get the source from the special selected partial method."
	
	| messageName |
	messageName := self selectedMethod.
	messageName ifNil: [ ^ '' ].
	messageName isSymbol ifTrue: [ ^ super selectedMessage ].
	^ messageName getSource "is a CSPartialMethod"! !

!CSContextBrowser methodsFor: 'ui' stamp: 'tfel 2/23/2012 14:41:58.076'!
selectedMessageCategoryName
	"Be tolerant, when messageCategoryListIndex is too big."
	
	self messageCategoryListIndex = 0 ifTrue: [ ^ nil ].
	^ self messageCategoryListIndex <= self messageCategoryList size
		ifTrue: [ self messageCategoryList at: self messageCategoryListIndex ]! !

!CSContextBrowser methodsFor: 'ui' stamp: 'ag 1/22/2012 23:06:18.598'!
selectedMessageName
	"The result of the super-call might be a CSPartialMethod, "
	
	^ super selectedMessageName ifNotNilDo: #asSymbol! !

!CSContextBrowser methodsFor: 'ui' stamp: 'ag 1/22/2012 23:00:29.567'!
selectedMethod
	"Access either the symbol or the CSPartialMethod currently selected."
	
	^ super selectedMessageName! !

!Symbol methodsFor: '*contexts2-core' stamp: 'ag 1/22/2012 14:51:21.558'!
asLayer
	"Return the first-class layer-object with the name represented by the receiver."
	
	^ CSLayer layerNamed: self! !

!Symbol methodsFor: '*contexts2-core' stamp: 'ag 1/22/2012 14:52:50.783'!
layerName
	"Also implemented on CSLayer, to transparently access the layer-name of a layer-object without knowing whether it actually is the object or the name-symbol already."
	
	^ self! !

!Symbol methodsFor: '*contexts2-core' stamp: 'ag 12/22/2011 15:26'!
withLayerDo: aBlock
	
	^ (Array with: self) withLayersDo: aBlock! !

!Symbol methodsFor: '*contexts2-core' stamp: 'ag 12/22/2011 15:26'!
withoutLayerDo: aBlock
	
	^ (Array with: self) withoutLayersDo: aBlock! !

!PackageInfo methodsFor: '*contexts2-core' stamp: 'tfel 2/23/2012 19:26:11.397'!
layeredMethods
	| coreMethods extensionMethods |
	coreMethods := self classesAndMetaClasses gather: [:class |
		class methodDict values select: [:method | method isContextSMethod]].
	extensionMethods := (self externalBehaviors gather: [:classOrTrait |
		(self extensionCategoriesForClass: classOrTrait)
			gather: [:cat | (classOrTrait organization listAtCategoryNamed: cat)
				collect: [:selector || method |
					method := classOrTrait methodDict at: selector.
					method isContextSMethod ifTrue: [method]]]])
		reject: #isNil.
	^ (coreMethods, extensionMethods) gather: [:method |
			method partialMethods collect: [:m |
				CSMethodReference onMethod: m]]! !

!PackageInfo methodsFor: '*contexts2-core-override-listing' stamp: 'tfel 2/23/2012 18:50:23.126'!
methods
	^ ((self extensionMethods, self coreMethods) select: [:method |
		method isValid
			and: [method isLocalSelector
			and: [method methodSymbol isDoIt not]]]), self layeredMethods! !

!Process methodsFor: '*contexts2-core' stamp: 'tfel 2/23/2012 14:49:20.42'!
csInfo
	"Only the contents of csInfo are accessed, thus no need for a setter in Process"
	
	^ self
		environmentAt: #csInfo
		ifAbsent: [self environmentAt: #csInfo put: CSProcessContextInformation new]! !

!Behavior methodsFor: '*contexts2-core' stamp: 'tfel 2/28/2012 15:05:58.093'!
allCompiledMethods

    ^ Array streamContents: [:stream |
        self withAllSuperclassesDo: [:aClass |
            aClass methodsDo: [:method | 
                stream nextPut: method]]]! !

!Behavior methodsFor: '*contexts2-core' stamp: 'ag 1/12/2012 12:55'!
allInvolvedLayers
	
	| layers |
	layers := Set new.
	methodDict values
		select: [ :each | each class == CSLayeredMethod ]
		thenDo: [ :each | layers addAll: each allInvolvedLayers ].
	^ layers! !

!Behavior methodsFor: '*contexts2-core-override-adding/removing methods' stamp: 'tfel 2/29/2012 10:39:13.524'!
basicAddSelector: selector withMethod: compiledMethod 
	"Add the message selector with the corresponding compiled method to the receiver's method dictionary. Do this without sending system change notifications"
	
	| oldMethodOrNil method |
	oldMethodOrNil := self lookupSelector: selector.
	"Empty symbols are used by Object >> withArgs:executeMethod: to temporarily store a method in the methodDict; avoid creating the LayeredMethod."
	"the >>respondsTo: check allows us to load this in one Mcz without the need for a postload package"
	compiledMethod methodClass: self; selector: selector.
	(selector notEmpty and: [self respondsTo: #contextSConversionOld:new:])
		ifFalse: [ method := compiledMethod ]
		ifTrue: [
			method := self
				contextSConversionOld: (self compiledMethodAt: selector ifAbsent: [ nil ])
				new: compiledMethod.
			method methodClass: self; selector: selector ].
	self methodDict at: selector put: method.
	
	"Now flush Squeak's method cache, either by selector or by method"
	oldMethodOrNil == nil ifFalse: [oldMethodOrNil flushCache].
	selector flushCache.! !

!Behavior methodsFor: '*contexts2-core' stamp: 'tfel 2/29/2012 11:12:11.475'!
contextSConversionOld: old new: new
	"old can be nil or a CompiledMethod or a CSLayeredMethod. new can be a CompiledMethod or a CSPartialMethod."
	| layeredMethod |
	"During bootstrap #asPartialMethod may not work"
	(new respondsTo: #asPartialMethod) ifFalse: [^ new].
	layeredMethod := self layeredMethodFrom: old.
	layeredMethod addPartialMethod: new asPartialMethod.
	^ layeredMethod isLayered
		ifTrue: [ layeredMethod ]
		ifFalse: [ new ] "use the CompiledMethod, if LayeredMethod would be useless."! !

!Behavior methodsFor: '*contexts2-core' stamp: 'ag 12/11/2011 16:43'!
cs2FileOut
	
	self cs2FileOutOn: (FileStream forceNewFileNamed: self name asString, '.st').! !

!Behavior methodsFor: '*contexts2-core' stamp: 'ag 12/11/2011 09:59'!
cs2FileOutCompiledMethod: compiledMethod called: aSelector of: aClass on: aStream
	
	self
		cs2FileOutMultiPart: 
			{ 
			aClass name, ' methodsFor: ''', (aClass organization categoryOfElement: aSelector), ''''.
			compiledMethod getSource
			}
		on: aStream.! !

!Behavior methodsFor: '*contexts2-core' stamp: 'ag 12/20/2011 20:33'!
cs2FileOutMethod: meth called: aSelector of: aClass on: stream
	
	(meth isKindOf: CSLayeredMethod)
		ifTrue: [ meth partialMethods do: [ :partialMethod | self cs2FileOutCompiledMethod: partialMethod method called: aSelector of: aClass on: stream ] ]
		ifFalse: [ self cs2FileOutCompiledMethod: meth called: aSelector of: aClass on: stream ]! !

!Behavior methodsFor: '*contexts2-core' stamp: 'ag 12/11/2011 09:57'!
cs2FileOutMultiPart: aCollectionOfStrings on: stream
	
	stream nextPut: $!!; cr.
	aCollectionOfStrings
		do: [ :str | self cs2FileOutPart: str on: stream ].
	stream nextPut: $!!; cr.! !

!Behavior methodsFor: '*contexts2-core' stamp: 'ag 12/11/2011 10:07'!
cs2FileOutOn: stream
	
	self cs2FileOutPart: self definition on: stream.
	self methodDict keysAndValuesDo: [ :selector :meth |
		self cs2FileOutMethod: meth called: selector of: self on: stream ].! !

!Behavior methodsFor: '*contexts2-core' stamp: 'ag 12/11/2011 09:56'!
cs2FileOutPart: aString on: stream
	
	| tokens |
	tokens := aString asString findTokens: '!!' keep: '!!'.
	tokens do: [ :str |
		stream nextPutAll: 
			(str = '!!'
				ifTrue: [ '!!!!' ]
				ifFalse: [ str ]) ].
	stream nextPut: $!!; cr; cr.! !

!Behavior methodsFor: '*contexts2-core' stamp: 'tfel 2/29/2012 11:07:52.838'!
layeredMethodFrom: oldMethod
	"oldMethod can be nil, a CompiledMethod or a CSLayeredMethod"
	oldMethod ifNil: [^ CSLayeredMethod new].
	^ oldMethod isCompiledMethod
		ifTrue: [CSLayeredMethod new
					addPartialMethod: oldMethod asPartialMethod;
					yourself]
		ifFalse: [oldMethod]! !

Object subclass: #CSLayer
	instanceVariableNames: 'layerName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ContextS2-Core'!

!CSLayer class methodsFor: 'private' stamp: 'ag 12/23/2011 14:22'!
checkIsSubclass
	
	self == CSLayer ifTrue: [ self error: 'Cannot access singleton-instance of CSLayer itself.' ].! !

!CSLayer class methodsFor: 'private' stamp: 'ag 12/23/2011 14:33'!
getGlobalInstanceOf: aClass withName: aSymbol
	
	^ (Smalltalk includesKey: aSymbol)
		ifTrue: [ | layer |
			layer := Smalltalk at: aSymbol.
			(layer isKindOf: CSLayer)
				ifTrue: [
					layer class == aClass
						ifTrue: [ layer ]
						ifFalse: [ "Delete obsolete CSLayer instances. Kind of a side case, but keeps the system clean."
							Smalltalk removeKey: aSymbol.
							self getGlobalInstanceOf: aClass withName: aSymbol ] ]
				ifFalse: [ self error: 'Global ', aSymbol, ' already exists, cannot create layer with that name.' ] ]
		ifFalse: [
			Smalltalk at: aSymbol put: (aClass basicNew initialize: aSymbol; yourself) ] "Avoid #new"! !

!CSLayer class methodsFor: 'subclassed layers' stamp: 'ag 12/22/2011 16:22'!
globalLayerName
	"The name of the layer to be accessed as a global variable. Is accessible as soon as instance has been invoked."
	
	self subclassResponsibility.! !

!CSLayer class methodsFor: 'subclassed layers' stamp: 'ag 12/23/2011 14:29'!
instance
	"Singleton-pattern with registration in the global namespace.
	Only for subclasses of CSLayer, CSLayer itself has multiple 'anonymous' instances (which are also registered globally)"
	
	self checkIsSubclass.
	^ self
		getGlobalInstanceOf: self
		withName: self globalLayerName! !

!CSLayer class methodsFor: 'layers' stamp: 'ag 12/23/2011 14:29'!
layerNamed: aSymbol
	"Return the CSLayer-instance associated with aSymbol.
	If a subclass with the according globalLayerName exists, it's singleton-instance is returned.
	Else, an anonymous instance of CSLayer is returned. These objects are created lazily, tracked in a dictionary and freed when become obsolet.e."
	
	^ (self subclassedLayerNamed: aSymbol)
		ifNotNilDo: #instance
		ifNil: [
			self
				getGlobalInstanceOf: CSLayer
				withName: aSymbol ]! !

!CSLayer class methodsFor: 'private' stamp: 'ag 12/22/2011 15:54'!
new
	
	^ self error: 'Dont create instances of CSLayer (or subclasses) manually.'! !

!CSLayer class methodsFor: 'private' stamp: 'ag 12/23/2011 14:23'!
pvtCreateNamed: aSymbol
	"Only internal, for controlled instance creation."
	
	^ self basicNew initialize: aSymbol; yourself! !

!CSLayer class methodsFor: 'private' stamp: 'ag 12/23/2011 14:32'!
removeGlobalInstanceNamed: aSymbol
	
	^ ((Smalltalk at: aSymbol ifAbsent: [ ^ nil ]) isKindOf: CSLayer)
		ifTrue: [ Smalltalk removeKey: aSymbol ]
		ifFalse: [ self error: 'Global ', aSymbol, ' is not an instance of CSLayer' ]! !

!CSLayer class methodsFor: 'subclassed layers' stamp: 'ag 12/23/2011 14:26'!
resetInstance
	"Reset the layer-instance of this class. Only applicable for subclasses."
	
	self checkIsSubclass.
	^ self removeGlobalInstanceNamed: self globalLayerName! !

!CSLayer class methodsFor: 'layers' stamp: 'ag 12/23/2011 14:26'!
resetLayer: aSymbol
	"Resets the global layer-instance named by aSymbol. Takes layers with their own subclass and anonymous layers into account."
	
	^ (self subclassedLayerNamed: aSymbol)
		ifNotNilDo: #resetInstance
		ifNil: [ self removeGlobalInstanceNamed: aSymbol ]! !

!CSLayer class methodsFor: 'private' stamp: 'ag 12/23/2011 14:11'!
subclassedLayerNamed: aSymbol
	
	^ self allSubclasses
		detect: [ :cl | cl globalLayerName = aSymbol ]
		ifNone: [ nil ]! !

!CSLayer methodsFor: 'queries' stamp: 'ag 12/23/2011 15:14'!
allContainingLayeredMethods
	
	^ CSLayeredMethod allInstances
		select: [ :method |
			method partialMethods
				anySatisfy: [ :partial | partial layers includes: self layerName ] ]! !

!CSLayer methodsFor: 'queries' stamp: 'ag 12/23/2011 15:16'!
allContainingMethods
	"Returns a collection of associations from class to selector."
	
	^ self allContainingLayeredMethods
		collect: [ :layeredMethod | layeredMethod methodClass -> layeredMethod selector ]! !

!CSLayer methodsFor: 'queries' stamp: 'ag 12/23/2011 15:52'!
allContainingPartialMethods
	
	| result |
	result := OrderedCollection new.
	self allContainingLayeredMethods
		do: [ :layeredMethod |
			result addAll:
				(layeredMethod partialMethods
					select: [ :partialMethod | partialMethod layers includes: self layerName ]) ].
	^ result! !

!CSLayer methodsFor: 'queries' stamp: 'ag 1/22/2012 14:52:09.315'!
asLayer
	"Also implemented on Symbol, to transparently access this first-class layer object."
	
	^ self! !

!CSLayer methodsFor: 'initialization' stamp: 'ag 12/22/2011 17:18'!
initialize
	
	self error: 'Please use the #initialize: message with the layer-name as parameter.'.! !

!CSLayer methodsFor: 'initialization' stamp: 'ag 12/22/2011 17:18'!
initialize: aSymbol
	
	super initialize.
	layerName := aSymbol.! !

!CSLayer methodsFor: 'accessing' stamp: 'ag 12/22/2011 17:16'!
layerName
	
	^ layerName! !

!CSLayer methodsFor: 'layer handling' stamp: 'ag 12/23/2011 15:33'!
reset
	"Reset the layer-instance. Next time the layer will be accessed, it will result in another object."
	
	^ self class resetLayer: self layerName! !

!CSLayer methodsFor: 'layer handling' stamp: 'ag 1/27/2012 16:49:15.994'!
withLayerDo: aBlock
	"Pass the call on to the name-symbol. TODO in future, the functionality should be replaced here, in the Layer-class."
	
	^ self layerName withLayerDo: aBlock! !

Object subclass: #CSMethodObject
	instanceVariableNames: 'methodClass selector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ContextS2-Core'!

CSMethodObject subclass: #CSLayeredMethod
	instanceVariableNames: 'baseMethod partialMethods'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ContextS2-Core'!

!CSLayeredMethod class methodsFor: 'consistency' stamp: 'tfel 2/29/2012 10:16:13.666'!
checkConsistencyOfInstances
	"self checkConsistencyOfInstances"
	
	super checkConsistencyOfInstances.
	[ CSLayeredMethod allInstances
		allSatisfy: [ :method |
			method partialMethods class == OrderedCollection ]
	] assert.

	[ CSLayeredMethod allInstances
		allSatisfy: [ :method |
			(method isLayered) or: [
			(method methodClass methodDict includes: method) not] ]
	] assert.! !

!CSLayeredMethod class methodsFor: 'consistency' stamp: 'ag 1/30/2012 13:05:45.269'!
fixConsistencyOfInstances
	"self fixConsistencyOfInstances"
	
	super fixConsistencyOfInstances.
	[ CSLayeredMethod allInstances
		do: [ :method | "Replace the collection-instance in the CSLayeredMethod-instance by a Set."
			method setPartialMethods: (OrderedCollection withAll: method partialMethods) ].
	] on: Error do: [ :e | Transcript cr; show: 'Fixing partialMethods-collection of all subinstances of CSLayeredMethod failed: ', e ].
	
	[ CSLayeredMethod allInstances
		do: [ :method | "Replace CSLayeredMethods, that do not actually include layered behaviour, by simple CompiledMethods."
			method isLayered ifFalse: [ method becomeCompiledMethod ] ].
	] on: Error do: [ :e | Transcript cr; show: 'Turning all non-layered subinstances of CSLayeredMethod into CompiledMethods failed: ', e ].! !

!CSLayeredMethod class methodsFor: 'constants' stamp: 'ag 12/20/2011 21:39'!
pragmaKeyword
	
	^ #layer:! !

!CSLayeredMethod class methodsFor: 'constants' stamp: 'ag 12/23/2011 16:08'!
proceedSelectorBase
	
	^ #proceed! !

!CSLayeredMethod methodsFor: 'accessing' stamp: 'ag 12/17/2011 21:50'!
activeLayers
	
	^ Processor activeProcess csInfo activeLayers! !

!CSLayeredMethod methodsFor: 'accessing' stamp: 'tfel 2/28/2012 15:45:04.655'!
addPartialMethod: aPartialMethod

	aPartialMethod ifNil: [ ^ self ].
	aPartialMethod isBase
		ifTrue: [ baseMethod := aPartialMethod ]
		ifFalse: [
			(self hasConflictsWithPartialMethod: aPartialMethod) ifFalse: [
				self partialMethods
					remove: aPartialMethod ifAbsent: [ "nothing" ];
					add: aPartialMethod ]].! !

!CSLayeredMethod methodsFor: 'accessing' stamp: 'ag 1/12/2012 13:00'!
allInvolvedLayers
	
	| layers |
	layers := Set new.
	self partialMethods
		do: [ :each | layers addAll: each layers ].
	^ layers! !

!CSLayeredMethod methodsFor: 'queries' stamp: 'ag 1/30/2012 16:30:13.142'!
allPartialMethods
	"Return the partial methods in this layered method, inclusive the base-method.
	Returned collection may not be modified, as the internal partial-method-collection may be returned without copying."
	
	^ self includesBaseMethod
		ifTrue: [ self partialMethods copyWith: baseMethod ]
		ifFalse: [ self partialMethods ]! !

!CSLayeredMethod methodsFor: 'accessing' stamp: 'ag 1/30/2012 16:33:16.835'!
baseMethod
	"This does NOT return just the instance field, but unpacks the CompiledMethod from the CSPartialMethod."
	
	^ baseMethod ifNotNilDo: #method! !

!CSLayeredMethod methodsFor: 'accessing' stamp: 'ag 1/30/2012 16:28:31.056'!
basePartialMethod
	
	^ baseMethod! !

!CSLayeredMethod methodsFor: 'queries' stamp: 'ag 1/30/2012 13:39:48.601'!
becomeCompiledMethod
	"This restores the consistency of this object, if it is not layered anymore. First check, if the receiver is still inside its class' methodDict.
	Then, if we have any partial method left, put it's CompiledMethod in the methodDict; else, just delete the slot.
	If this is called while #isLayered returns true, some of the partial methods will just get lost."
	
	(self methodClass methodDict at: self selector ifAbsent: [ nil ]) == self ifFalse: [ ^ self ].
	self includesAnyPartialMethod ifFalse: [ self methodClass methodDict removeKey: self selector ifAbsent: [ "nothing" ] ].
	self compiledMethodReplacement ifNotNilDo: [ :compiledMethod |
		self methodClass methodDict
			at: self selector
			put: compiledMethod ].! !

!CSLayeredMethod methodsFor: 'accessing' stamp: 'ag 1/30/2012 13:39:21.924'!
compiledMethodReplacement
	
	^ self includesAnyPartialMethod ifTrue: [
		self includesBaseMethod
			ifTrue: [ self baseMethod ]
			ifFalse: [ self partialMethods anyOne method ] ]! !

!CSLayeredMethod methodsFor: 'accessing' stamp: 'ag 1/20/2012 22:44:40.615'!
hasBaseMethod
	
	^ self baseMethod notNil! !

!CSLayeredMethod methodsFor: 'accessing' stamp: 'tfel 2/28/2012 15:48:52.912'!
hasConflictsWithPartialMethod: aPartialMethod

	^ self partialMethods
			detect: [:method | method layers ~= aPartialMethod layers and: [
								method layers includesAnyOf: aPartialMethod layers]]
			ifFound: [:method |
				((self confirm: 'This will override another method that defines the layers ',
								(method layers intersection: aPartialMethod layers))
					ifTrue: [self partialMethods remove: method. false]
					ifFalse: [true])]
			ifNone: [false]! !

!CSLayeredMethod methodsFor: 'accessing' stamp: 'ag 1/23/2012 08:35:04.233'!
includesAnyPartialMethod
	"Return true, if any behaviour is still left in this layered method."
	
	^ self includesBaseMethod | self partialMethods notEmpty! !

!CSLayeredMethod methodsFor: 'accessing' stamp: 'ag 1/30/2012 16:31:49.261'!
includesBaseMethod
	
	^ self baseMethod notNil! !

!CSLayeredMethod methodsFor: 'accessing' stamp: 'ag 1/22/2012 17:24:33.091'!
isLayered
	"Returns true, if this layered method currently does anything different than a simple CompiledMethod.
	If false, this layered method should be replaced by an equivalent CompiledMethod.
	This is the case, if there are no partial methods."
	
	^ self partialMethods notEmpty! !

!CSLayeredMethod methodsFor: 'accessing' stamp: 'ag 1/30/2012 16:30:48.529'!
method
	"Use base method, if possible, else use any partial method. Return nil, if not a single partial method is present."
	
	^ self includesBaseMethod
		ifTrue: [ self baseMethod ]
		ifFalse: [ self partialMethods ifNotEmpty: [ self partialMethods anyOne ] ifEmpty: [ nil ] ]! !

!CSLayeredMethod methodsFor: 'executing' stamp: 'tfel 2/28/2012 15:58:15.156'!
methodChainFor: activeLayers
	"Chain together all partial methods that are currently active. Chain is ordered in the way the layers have been activated."
	
	| methodChain applicableMethods layeredMethods |
	methodChain := OrderedCollection new.
	applicableMethods := self methodClass allCompiledMethods
								select: [:method | method selector = self selector].
	layeredMethods := (applicableMethods
							select: #isContextSMethod)
							gather: [:layeredMethod | layeredMethod partialMethods].
	(activeLayers do: [ :layer |
		(layeredMethods
			detect: [:method | method layers includes: layer]
			ifFound: [:method | methodChain addIfNotPresent: method]
			ifNone: [])]).
	(applicableMethods
		detect: [:method | method isContextSMethod not or: [
							method includesBaseMethod]]
		ifFound: [:method | methodChain add: (method isContextSMethod 
												ifTrue: [method baseMethod]
												ifFalse: [method])]
		ifNone: ["doesn't matter"]).
	^ methodChain! !

!CSLayeredMethod methodsFor: 'accessing' stamp: 'ag 1/27/2012 17:07:35.681'!
partialMethods
	
	^ partialMethods ifNil: [ partialMethods := OrderedCollection new ]! !

!CSLayeredMethod methodsFor: 'printing' stamp: 'ag 12/23/2011 15:25'!
printOn: aStream
	
	super printOn: aStream.
	aStream
		nextPutAll: ', ';
		nextPutAll: (self partialMethods size + (self includesBaseMethod ifTrue: [ 1 ] ifFalse: [ 0 ])) asString;
		nextPutAll: ' version(s))'.! !

!CSLayeredMethod methodsFor: 'executing' stamp: 'ag 12/25/2011 18:07'!
proceedSelector
	
	| selector |
	selector := self class proceedSelectorBase.
	self selector isUnary ifTrue: [ ^ selector asSymbol ].
	selector := selector, #':'.
	self selector isInfix ifTrue: [ ^ selector asSymbol ].
	(self selector occurrencesOf: $:) - 1
		timesRepeat: [ selector := selector, #with: ].
	^ selector asSymbol! !

!CSLayeredMethod methodsFor: 'executing' stamp: 'nw 1/22/2012 15:26'!
registerProceed: aReceiver
	"If no ProceedMethod-instance was in the method-dictionary, create one and insert it. If #proceed: was implemented otherwise, throw an error."
	
	(aReceiver class methodDict
		includesKey: self proceedSelector)
		ifTrue: [ | partialMethod |
			partialMethod := aReceiver class methodDict at: self proceedSelector.
			partialMethod class == CSProceedMethod
				ifFalse: [ self error: 'ContextS may not be used with classes implementing #', self proceedSelector ].
			partialMethod incrementRunningLayeredMethods ]
		ifFalse: [
			self flag: #ag. "Access to methodDict should be synchronized on a semaphore. See also #unregisterProceed:"
			aReceiver class methodDict
				at: self proceedSelector
				put: (CSProceedMethod selector: self proceedSelector class: self class) ]. "Default runningLayeredMethods is 1."! !

!CSLayeredMethod methodsFor: 'queries' stamp: 'ag 1/30/2012 16:32:19.148'!
removePartialMethod: aPartialMethod
	
	aPartialMethod method == self baseMethod
		ifTrue: [ baseMethod := nil ]
		ifFalse: [ self partialMethods remove: aPartialMethod ifAbsent: [] ].
	self isLayered ifFalse: [ self becomeCompiledMethod ].! !

!CSLayeredMethod methodsFor: 'accessing' stamp: 'ag 1/22/2012 16:48:23.532'!
setPartialMethods: aCollection
	"Only for internal use."
	
	partialMethods := aCollection! !

!CSLayeredMethod methodsFor: 'executing' stamp: 'ag 12/23/2011 16:10'!
unregisterProceed: aReceiver
	"We are finished using the ProceedMethod instance. Decrease its counter and remove it from the method-dictionary if necessary."
	
	(aReceiver class methodDict
		at: self proceedSelector
		ifAbsent: [ ^ self ])
		ifNotNilDo: [ :proceedMethod |
			(proceedMethod isKindOf: CSProceedMethod)
				ifTrue: [
					proceedMethod decrementRunningLayeredMethods.
					proceedMethod runningLayeredMethods <= 0
						ifTrue: [ 
							self flag: #ag. "Access to methodDict should be synchronized on a semaphore. See also #registerProceed:"
							aReceiver class methodDict removeKey: self proceedSelector ] ] ].! !

!CSLayeredMethod methodsFor: 'executing' stamp: 'tfel 2/28/2012 15:26:06.403'!
valueWithReceiver: aReceiver arguments: anArray
	"Memorize old method-chain, compute a new one, kick off the execution and restore the old state afterwards."
	
	| oldProcessState methodChain |
	methodChain := self methodChainFor: self activeLayers.
	methodChain ifEmpty: ["Not a single method to execute..."
		^ aReceiver
			doesNotUnderstand: (Message selector: self selector arguments: anArray) ].
	self registerProceed: aReceiver.
	oldProcessState := Processor activeProcess csInfo saveState.
	Processor activeProcess csInfo methodChain: methodChain.
	^ [ methodChain first valueWithReceiver: aReceiver arguments: anArray ]
		ensure: [
			self unregisterProceed: aReceiver.
			Processor activeProcess csInfo restoreStateFrom: oldProcessState ]! !

!CSMethodObject class methodsFor: 'consistency' stamp: 'ag 1/27/2012 17:52:04.164'!
checkAllInstances
	
	Smalltalk garbageCollect. "As the check/fix methods use allInstances/allSubInstances, we should try to collect all obsolete instances."
	^ CSMethodObject allSubclassesDo: #checkConsistencyOfInstances! !

!CSMethodObject class methodsFor: 'consistency' stamp: 'ag 1/30/2012 12:52:38.416'!
checkConsistencyOfInstances
	"self checkConsistencyOfInstances"
	
	[ self allInstances
		noneSatisfy: [ :method |
			method selector isNil or: [ method selector isEmpty ]
			or: [ method methodClass isNil ] ]
	] assert.! !

!CSMethodObject class methodsFor: 'consistency' stamp: 'ag 1/27/2012 17:51:48.267'!
fixAllInstances
	
	Smalltalk garbageCollect. "As the check/fix methods use allInstances/allSubInstances, we should try to collect all obsolete instances."
	^ CSMethodObject allSubclassesDo: #fixConsistencyOfInstances! !

!CSMethodObject class methodsFor: 'consistency' stamp: 'ag 1/30/2012 12:55:08.897'!
fixConsistencyOfInstances
	"self fixConsistencyOfInstances"
	"Sets the selector of all method-object instances to the selector of their underlying CompiledMethod."
	
	[ self allInstances do: [ :methodObject |
		methodObject selector: methodObject method selector ]
	] on: Error do: [ :e | Transcript cr; show: 'Fixing selectors on all instances of ', self, ' failed: ', e ].
	
	[ self allInstances do: [ :methodObject |
		methodObject selector ifEmpty: [ methodObject methodClass methodDict removeKey: #'' ifAbsent: [] ] ]
	] on: Error do: [ :e | Transcript cr; show: 'Deleting instances of ', self, ' with empty selector failed: ', e ]! !

!CSMethodObject class methodsFor: 'instance creation' stamp: 'ag 12/20/2011 20:12'!
selector: aSymbol class: aClass
	
	^ self new
		selector: aSymbol;
		methodClass: aClass;
		yourself! !

!CSMethodObject methodsFor: 'converting' stamp: 'tfel 2/23/2012 17:45:41.4'!
asSymbol
	
	^ self printString asSymbol! !

!CSMethodObject methodsFor: 'accessing' stamp: 'ag 1/19/2012 19:01:38.067'!
at: aNumber
	
	^ self method at: aNumber! !

!CSMethodObject methodsFor: 'executing' stamp: 'ag 12/20/2011 21:19'!
doesNotUnderstand: aMessage
	"Required for all meta messages that need information from the original method."
	
	^ self method
		ifNil: [ self ] "Cannot pass the message on, simulate behavior of an empty method."
		ifNotNilDo: [ :msg | msg perform: aMessage selector withArguments: aMessage arguments ]! !

!CSMethodObject methodsFor: 'literals' stamp: 'tfel 2/29/2012 10:05:23.663'!
hasLiteral: literal

	self method ifNil: [^ false].
	^ self method hasLiteral: literal! !

!CSMethodObject methodsFor: 'literals' stamp: 'tfel 2/29/2012 10:05:35.373'!
hasLiteralThorough: literal

	self method ifNil: [^ false].
	^ self method hasLiteralThorough: literal! !

!CSMethodObject methodsFor: 'testing' stamp: 'tfel 2/23/2012 17:03:30.372'!
isContextSMethod

	^ true! !

!CSMethodObject methodsFor: 'executing' stamp: 'ag 12/20/2011 19:58'!
method
	"Subclasses have to provide real CompiledMethod-objects as substitute for arbitrary messages. If nil is returned, these messages are dropped."
	
	^ self subclassResponsibility! !

!CSMethodObject methodsFor: 'accessing' stamp: 'ag 12/20/2011 20:11'!
methodClass
	"Answer the value of methodClass"

	^ methodClass! !

!CSMethodObject methodsFor: 'accessing' stamp: 'ag 12/20/2011 20:11'!
methodClass: anObject
	"Set the value of methodClass"

	methodClass := anObject! !

!CSMethodObject methodsFor: 'printing' stamp: 'ag 12/23/2011 15:26'!
printOn: aStream
	
	aStream
		nextPutAll: self class name;
		nextPutAll: '(';
		nextPutAll: (self methodClass name ifNil: [ '' ]);
		nextPutAll: '>>';
		nextPutAll: (self selector ifNil: [ #'' ]).! !

!CSMethodObject methodsFor: 'executing' stamp: 'ag 12/16/2011 11:45'!
run: aSelector with: anArrayOfObjects in: aReceiver
	
	^ self valueWithReceiver: aReceiver arguments: anArrayOfObjects! !

!CSMethodObject methodsFor: 'accessing' stamp: 'ag 1/30/2012 16:36:22.041'!
selector
	"Answer the value of selector"

	^ selector! !

!CSMethodObject methodsFor: 'accessing' stamp: 'ag 1/31/2012 15:23:23.931'!
selector: anObject
	"Set the value of selector"
	
	selector := anObject! !

!CSMethodObject methodsFor: 'accessing' stamp: 'ag 1/19/2012 19:01:43.145'!
size
	
	^ self method size! !

!CSMethodObject methodsFor: 'executing' stamp: 'ag 12/20/2011 19:59'!
valueWithReceiver: aReceiver arguments: anArrayOfObjects
	"This message is sent, when the method-object has to execute itself."
	
	^ self subclassResponsibility! !

CSMethodObject subclass: #CSPartialMethod
	instanceVariableNames: 'method layers'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ContextS2-Core'!

!CSPartialMethod class methodsFor: 'consistency' stamp: 'ag 1/30/2012 13:08:04.048'!
checkConsistencyOfInstances
	
	super checkConsistencyOfInstances.
	
	[ CSPartialMethod allInstances noneSatisfy: [ :method |
		(method methodClass methodDict at: method selector ifAbsent: [])
			== method ]
	] assert.
	
	"TODO another helpfull check would be to check, that no partial methods return true on the #isBase message..."
	"TODO Yet another check would be to check, that instances of partialMEthod, that return true from #isBase really include the layer: pragma in their CompiledMethod"! !

!CSPartialMethod class methodsFor: 'consistency' stamp: 'ag 1/27/2012 17:48:46.172'!
fixConsistencyOfInstances
	
	super fixConsistencyOfInstances.
	
	[ CSPartialMethod allSubInstances do: [ :methodObject |
		(methodObject methodClass methodDict at: methodObject selector ifAbsent: []) == methodObject
			ifTrue: [
				methodObject methodClass methodDict removeKey: methodObject selector ifAbsent: [] ] ]
	] on: Error do: [ :e | Transcript cr; show: 'Deleting all instances of CSPartialMethod from method-dictionaries failed: ', e ].
	
	[ CSLayeredMethod allSubInstances do: [ :methodObject |
		methodObject setPartialMethods: (methodObject partialMethods reject: [ :e | e selector isEmpty ])]
	] on: Error do: [ :e | Transcript cr; show: 'Deleting all instances of CSPartialMethod with empty selectors from their respective layered method failed: ', e ].! !

!CSPartialMethod class methodsFor: 'instance creation' stamp: 'ag 12/23/2011 16:49'!
on: aLayer forMethod: aMethod
	
	^ (self
		selector: aMethod selector
		class: aMethod methodClass)
		method: aMethod;
		layers: aLayer;
		yourself! !

!CSPartialMethod methodsFor: 'comparing' stamp: 'ag 1/30/2012 13:01:04.742'!
= anotherPartialMethod
	
	^ anotherPartialMethod class == self class
		and: [ (layers = anotherPartialMethod internalLayers) & (self selector = anotherPartialMethod selector) & (self methodClass = anotherPartialMethod methodClass) ]! !

!CSPartialMethod methodsFor: 'comparing' stamp: 'ag 1/22/2012 17:16:13.838'!
hash
	
	^ layers hash + self selector hash + self class hash! !

!CSPartialMethod methodsFor: 'accessing' stamp: 'ag 1/22/2012 23:11:26.719'!
internalLayers
	"Returns the instance-variable without further checks. For internal use only."
	
	^ layers! !

!CSPartialMethod methodsFor: 'accessing' stamp: 'ag 1/22/2012 17:17:06.694'!
isBase
	
	^ layers isNil! !

!CSPartialMethod methodsFor: 'accessing' stamp: 'ag 1/22/2012 17:04:45.476'!
layers
	"Answer the value of layer. If it is nil, return an empty array, but DONT set the instance-variable itself!!
	The instance variable layer being nil is a marker for this partial-method to be the base-method (see #isBase)."
	
	^ layers ifNil: [ #() ]! !

!CSPartialMethod methodsFor: 'accessing' stamp: 'ag 2/2/2012 10:34:56.272'!
layers: anObject
	"Set the value of layer"

	anObject ifNotNil: [ anObject do: [ :symbol | CSLayer layerNamed: symbol ] ].
	layers := anObject! !

!CSPartialMethod methodsFor: 'accessing' stamp: 'mp 6/13/2010 14:19'!
method
	"Answer the value of method"

	^ method! !

!CSPartialMethod methodsFor: 'accessing' stamp: 'mp 6/13/2010 14:19'!
method: anObject
	"Set the value of method"

	method := anObject! !

!CSPartialMethod methodsFor: 'printing' stamp: 'ag 12/23/2011 15:28'!
printLayersOn: stream
	
	self layers
		do: [ :each | stream nextPutAll: each ]
		separatedBy: [ stream nextPut: Character space ].! !

!CSPartialMethod methodsFor: 'printing' stamp: 'ag 1/22/2012 16:57:41.933'!
printOn: aStream
	
	aStream
		nextPutAll: selector;
		nextPutAll: ' ('.
	self isBase
		ifTrue: [ aStream nextPutAll: '<base>' ]
		ifFalse: [ self printLayersOn: aStream ].
	aStream nextPut: $).! !

!CSPartialMethod methodsFor: 'executing' stamp: 'ag 12/21/2011 00:09'!
valueWithReceiver: aReceiver arguments: anArray
	"Memorize in the process-state, that the current method has been started; execute the backing CompiledMethod in the receivers context."
	
	Processor activeProcess csInfo executedPartialMethod: self.
	^ [ aReceiver withArgs: anArray executeMethod: self method ]
		ensure: [ Processor activeProcess csInfo executedPartialMethod: nil ] "Allow garbage collection of this partial method"! !

CSMethodObject subclass: #CSProceedMethod
	instanceVariableNames: 'runningLayeredMethods'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ContextS2-Core'!

!CSProceedMethod class methodsFor: 'consistency' stamp: 'ag 1/30/2012 12:53:16.454'!
checkConsistencyOfInstances
	"self checkSystemConsistency"
	
	"super checkConsistencyOfInstances." "Super-version not applicable."
	[ CSProceedMethod allInstances
		noneSatisfy: [ :method |
			method methodClass methodDict includesKey: method selector ]
	] assert.! !

!CSProceedMethod class methodsFor: 'consistency' stamp: 'ag 1/23/2012 08:42:51.977'!
fixConsistencyOfInstances
	"self fixConsistencyOfInstances"
	"Remove all leftover instances of ProceedMethod in method-dictionaries of classes.
	ProceedMethod-instances are removed inside ensure:-blocks, but nevertheless sometimes some of them can stay somewhere in method-dictionaries..."
	
	| number |
	"super fixConsistencyOfInstances." "Super-version not applicable."
	number := 0.
	[ CSProceedMethod allInstancesDo: [ :proceed |
			(proceed methodClass methodDict includesKey: proceed selector)
				ifTrue: [
					number := number + 1.
					proceed methodClass methodDict removeKey: proceed selector ] ].
	] on: Error do: [ :e | Transcript cr; show: 'Removing all unused subinstances of CSProceedMethod failed: ', e ].
	^ number! !

!CSProceedMethod methodsFor: 'accessing' stamp: 'ag 12/20/2011 21:01'!
decrementRunningLayeredMethods
	
	runningLayeredMethods := self runningLayeredMethods - 1.! !

!CSProceedMethod methodsFor: 'accessing' stamp: 'ag 12/20/2011 21:09'!
incrementRunningLayeredMethods
	
	runningLayeredMethods := self runningLayeredMethods + 1.! !

!CSProceedMethod methodsFor: 'executing' stamp: 'ag 12/20/2011 20:58'!
method
	"There is no real CompiledMethod to pass incoming messages to."
	
	^ nil! !

!CSProceedMethod methodsFor: 'executing' stamp: 'ag 12/23/2011 15:55'!
nextPartialMethod
	"Based on executedPartialMethod, find the next method in the methodChain."
	
	| info index |
	info := Processor activeProcess csInfo.
	index := info methodChain indexOf: info executedPartialMethod.
	info methodChain size <= index ifTrue: [ ^ nil ]. "Method chain has been executed to the end."
	^ info methodChain at: index + 1! !

!CSProceedMethod methodsFor: 'printing' stamp: 'ag 12/23/2011 15:31'!
printOn: aStream
	
	super printOn: aStream.
	aStream
		nextPutAll: ', count ';
		nextPutAll: (runningLayeredMethods ifNil: [ 0 ]) asString;
		nextPut: $).! !

!CSProceedMethod methodsFor: 'accessing' stamp: 'ag 12/20/2011 21:11'!
runningLayeredMethods
	"The default value is 1 to account for the layered method that created this proceed-method."
	
	^ runningLayeredMethods ifNil: [ runningLayeredMethods := 1 ]! !

!CSProceedMethod methodsFor: 'executing' stamp: 'nw 1/19/2012 19:22'!
valueWithReceiver: aReceiver arguments: anArray
	
	^ self nextPartialMethod
		ifNil: [ aReceiver ] "No method to proceed to. Simulate behavior of an empty method. Either the base method has called proceed: or there is no base method."
		ifNotNilDo: [ :partialMethod |
			aReceiver withArgs: anArray executeMethod: partialMethod ]! !

Object subclass: #CSProcessContextInformation
	instanceVariableNames: 'activeLayers methodChain executedPartialMethod'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ContextS2-Core'!

!CSProcessContextInformation methodsFor: 'accessing' stamp: 'ag 12/20/2011 20:00'!
activeLayers
	^ activeLayers ifNil: [ activeLayers := #() ]! !

!CSProcessContextInformation methodsFor: 'accessing' stamp: 'nw 12/12/2011 17:05'!
activeLayers: layers
	activeLayers := layers! !

!CSProcessContextInformation methodsFor: 'accessing' stamp: 'nw 12/12/2011 17:06'!
executedPartialMethod
	^ executedPartialMethod! !

!CSProcessContextInformation methodsFor: 'accessing' stamp: 'nw 12/12/2011 17:06'!
executedPartialMethod: partialMethod
	executedPartialMethod := partialMethod! !

!CSProcessContextInformation methodsFor: 'accessing' stamp: 'ag 12/18/2011 11:13'!
methodChain
	
	^ methodChain! !

!CSProcessContextInformation methodsFor: 'accessing' stamp: 'ag 12/18/2011 11:13'!
methodChain: anOrderedCollection
	
	methodChain := anOrderedCollection.! !

!CSProcessContextInformation methodsFor: 'layer handling' stamp: 'ag 12/23/2011 15:48'!
restoreStateFrom: anArray
	"anArray must be the return value of an earlier #saveState call."
	
	self methodChain: anArray first.
	self executedPartialMethod: anArray second.! !

!CSProcessContextInformation methodsFor: 'layer handling' stamp: 'ag 12/23/2011 15:48'!
saveState
	
	^ { self methodChain. self executedPartialMethod }! !

!CSProcessContextInformation methodsFor: 'layer handling' stamp: 'ag 12/20/2011 23:54'!
withLayers: aCollection do: aBlock
	"Convenience method for layer-manipulation. Stores old layers and restores them after executing aBlock."
	
	| oldLayers |
	oldLayers := self activeLayers.
	self activeLayers: aCollection.
	^ aBlock ensure: [ self activeLayers: oldLayers ]! !

!Object methodsFor: '*contexts2-core' stamp: 'ag 12/23/2011 11:30'!
currentLayer
	"Return the first of the currentLayers for convenience.
	This should be the first activated layer that has lead to the activation of the last partial method."
	
	^ self currentLayers ifNotNilDo: #first! !

!Object methodsFor: '*contexts2-core' stamp: 'ag 12/23/2011 11:33'!
currentLayers
	"Return the layers of the current execution. If no method associated to any layer is on the execution stack, nil is returned.
	Else, the last partial method that has started execution determines the 'current layers'.
	These are the layers that have lead to the activation of this partial method (-> the subset of the currently active layers, that intersects with the partial methods layers).
	This returns (sub)instances of CSLayer.
	If a subclass of CSLayer with the right globalLayerName exists, its singleton-instance is used, else an anonymous instance of CSLayer."
	
	| info |
	info := Processor activeProcess csInfo.
	^ info executedPartialMethod
		ifNotNilDo: [ :partialMethod |
			((info activeLayers intersection: partialMethod layers)
				collect: [ :layername | CSLayer layerNamed: layername ])
					ifEmpty: [ nil ] ]! !

!Object methodsFor: '*contexts2-core' stamp: 'tfel 2/23/2012 17:05:44.781'!
isContextSMethod

	^ false! !

!MCMethodDefinition methodsFor: '*contexts2-core' stamp: 'tfel 2/23/2012 19:09:07.297'!
selector: aSymbol

	selector := aSymbol! !

!CompiledMethodWithNode class methodsFor: '*contexts2-core-override-instance creation' stamp: 'ag 12/20/2011 21:39'!
generateMethodFromNode: aMethodNode trailer: bytes
	
	| method |
	method := aMethodNode generate: bytes.
	(method pragmas anySatisfy: [:each | each keyword = CSLayeredMethod pragmaKeyword ])
		ifTrue: [method := method asPartialMethod].
	^ self method: method node: aMethodNode.! !
CSBrowser initialize!
